// File RTCG3D.cs --- calling generated methods that contain loops
// sestoft@dina.kvl.dk * 2002-09-18, 2005-10-05

using System;
using System.Reflection;
using System.Reflection.Emit;

class RTCG3 {
  public static void Main(String [] args) {
    int count = int.Parse(args[0]);
    // Build: public static void MyMethod1(int x) { ... }
    DynamicMethod methodBuilder1 =
      new DynamicMethod("MyMethod1", 
			typeof(void),
			new Type[] { typeof(int) },
			typeof(String).Module);
    {
      // Obtain an IL generator to build the method body
      ILGenerator ilg = methodBuilder1.GetILGenerator();
      ilg.EmitWriteLine("MyClass.MyMethod1() was called");
      Label start = ilg.DefineLabel();
      ilg.MarkLabel(start);
      ilg.Emit(OpCodes.Ldarg_0);
      ilg.Emit(OpCodes.Ldc_I4_1);
      ilg.Emit(OpCodes.Sub);
      ilg.Emit(OpCodes.Starg_S, 0);
      ilg.Emit(OpCodes.Ldarg_0);
      ilg.Emit(OpCodes.Brtrue, start);
      ilg.Emit(OpCodes.Ret);  
    }
    // Build: public static void MyMethod2(int x) { ... }
    DynamicMethod methodBuilder2 =
      new DynamicMethod("MyMethod2", 
			typeof(void),
			new Type[] { typeof(int) },
			typeof(String).Module);
    {
      // Obtain an IL generator to build the method body
      ILGenerator ilg = methodBuilder2.GetILGenerator();
      ilg.EmitWriteLine("MyClass.MyMethod2() was called");
      Label start = ilg.DefineLabel();
      ilg.Emit(OpCodes.Ldarg_0);
      ilg.MarkLabel(start);
      ilg.Emit(OpCodes.Ldc_I4_1);
      ilg.Emit(OpCodes.Sub);
      ilg.Emit(OpCodes.Dup);
      ilg.Emit(OpCodes.Brtrue, start);
      ilg.Emit(OpCodes.Pop);
      ilg.Emit(OpCodes.Ret);  
    }
    // Build: public static void MyMethod3(int x) { ... }
    DynamicMethod methodBuilder3 =
      new DynamicMethod("MyMethod3", 
			typeof(void),
			new Type[] { typeof(int) },
			typeof(String).Module);
    {
      // Obtain an IL generator to build the method body
      ILGenerator ilg = methodBuilder3.GetILGenerator();
      ilg.EmitWriteLine("MyClass.MyMethod3() was called");
      Label start = ilg.DefineLabel();
      ilg.DeclareLocal(typeof(int));
      ilg.Emit(OpCodes.Ldarg_0);
      ilg.Emit(OpCodes.Stloc_0);
      ilg.MarkLabel(start);     
      ilg.Emit(OpCodes.Ldloc_0); 
      ilg.Emit(OpCodes.Ldc_I4_1);
      ilg.Emit(OpCodes.Sub);
      ilg.Emit(OpCodes.Stloc_0);
      ilg.Emit(OpCodes.Ldloc_0); 
      ilg.Emit(OpCodes.Brtrue, start);
      ilg.Emit(OpCodes.Ret);  
    }
    Timer t1 = new Timer();
    I2V method1 = (I2V)methodBuilder1.CreateDelegate(typeof(I2V));
    method1(count);
    Console.WriteLine("Generated IL method 1: " + t1.Check() + " sec");
    Timer t2 = new Timer();
    I2V method2 = (I2V)methodBuilder2.CreateDelegate(typeof(I2V));
    method2(count);
    Console.WriteLine("Generated IL method 2: " + t2.Check() + " sec");
    Timer t3 = new Timer();
    I2V method3 = (I2V)methodBuilder3.CreateDelegate(typeof(I2V));
    method3(count);
    Console.WriteLine("Generated IL method 3: " + t3.Check() + " sec");
    Timer t4 = new Timer();
    YourMethod(count);
    Console.WriteLine("Compiled  C# method: " + t4.Check() + " sec");
  }

  public static void YourMethod(int n) {
    do { 
      n--;
    } while (n != 0);
  }

  /* CIL code for YourMethod, if compiledwith optimization (csc /o):

    .method public hidebysig static void  YourMethod(int32 n) cil managed
    {
      // Code size       9 (0x9)
      .maxstack  2
      IL_0000:  ldarg.0
      IL_0001:  ldc.i4.1
      IL_0002:  sub
      IL_0003:  starg.s    n
      IL_0005:  ldarg.0
      IL_0006:  brtrue.s   IL_0000

      IL_0008:  ret
    } // end of method Test::YourMethod
  */
}

public delegate void I2V(int x);

// Crude timing utility ----------------------------------------
   
public class Timer {
  private DateTime start;

  public Timer() {
    start = DateTime.Now;
  }

  public double Check() {
    TimeSpan dur = DateTime.Now - start;
    return dur.TotalSeconds;
  }
}

// This weirdness from MS .NET 1.0 has gone away in .NET 2.0 beta 2:

// RTCG method MyMethod1 takes 3.9 sec for 1 bn iterations if
// MyMethod3 is not called, and only 2.5 sec if MyMethod3 is called,
// even if it is called only *after* MyMethod1 has run to completion.
// More precisely, if the compiler can statically determine that
// MyMethod3 *cannot* be called, then MyMethod1 is slow.  If MyMethod3
// *might* be called (but even if it isn't) then MyMethod1 is fast!??!
