// File RTCG4D.cs --- the power example, for CLR 2.0
// sestoft@dina.kvl.dk * 2002-10-02, 2005-04-12

// To allow this to run off a network drive, set Start | Settings |
// Control Panel | Administration | Microsoft .Net Framework 2.0
// Wizards | Adjust .net Security | Make changes to this computer |
// Local Intranet | Full trust.  Phew!

// CLR 1.1:

// A reflective call to specialized: 25.0 sec
// A delegate call to specialized: 0.671
// Static call to general compiled: 0.751
// Delegate call to general compiled: 1.06

// Reflective call to specialized generated method: 15,2919888 sec
// Delegate call to specialized generated method: 0,100144 sec
// Static call to general compiled method: 0,1301872 sec
// Interface call to general compiled method: 0,1402016 sec
// Delegate call to general compiled method: 0,1301872 sec

using System;
using System.Reflection;
using System.Reflection.Emit;

class RTCG4 : IIntInt2Int {
  public static void Main(String [] args) {
    int count = int.Parse(args[0]);
    int n = 16;
    // Build: public static int MyPower(int x) { ... }
    DynamicMethod methodBuilder =
      new DynamicMethod("MyPower", 
                        typeof(int),
                        new Type[] { typeof(int) },
                        typeof(String).Module);
    // Obtain an IL generator to build the method body
    ILGenerator ilg = methodBuilder.GetILGenerator();
    PowerGen(ilg, n);
    
    Int2Int power_n_dlg 
      = (Int2Int)(methodBuilder.CreateDelegate(typeof(Int2Int)));

    int res = 0;

    Timer t0 = new Timer();
    object[] arglist = new object[] { 3 };
    for (int i=count; i>0; i--)
      res = (int)methodBuilder.Invoke(null, arglist);      
    Console.WriteLine("Reflective call to specialized generated method: " 
                      + t0.Check() + " sec");
    Console.WriteLine(res);

    Timer t2 = new Timer();
    for (int i=count; i>0; i--)
      res = power_n_dlg(3);
    Console.WriteLine("Delegate call to specialized generated method: " 
                      + t2.Check() + " sec");
    Console.WriteLine(res);

    Timer t3 = new Timer();
    for (int i=count; i>0; i--)
      res = StaticPower(n, 3);
    Console.WriteLine("Static call to general compiled method: " 
                      + t3.Check() + " sec");
    Console.WriteLine(res);
    
    Timer t4 = new Timer();
    IIntInt2Int power = (IIntInt2Int)new RTCG4();
    for (int i=count; i>0; i--)
      res = power.Power(n, 3);
    Console.WriteLine("Interface call to general compiled method: " 
                      + t4.Check() + " sec");
    Console.WriteLine(res);

    Timer t5 = new Timer();
    IntInt2Int powerDlg = new IntInt2Int(StaticPower);
    for (int i=count; i>0; i--)
      res = powerDlg(n, 3);
    Console.WriteLine("Delegate call to general compiled method: " 
                      + t5.Check() + " sec");
    Console.WriteLine(res);
  }

  public int Power(int n, int x) {
    int p;
    p = 1;
    while (n > 0) {
      if (n % 2 == 0) 
        { x = x * x; n = n / 2; }
      else 
        { p = p * x; n = n - 1; }
    }
    return p;
  }

  public static int StaticPower(int n, int x) {
    int p;
    p = 1;
    while (n > 0) {
      if (n % 2 == 0) 
        { x = x * x; n = n / 2; }
      else 
        { p = p * x; n = n - 1; }
    }
    return p;
  }

  public static void PowerGen(ILGenerator ilg, int n) {
    ilg.DeclareLocal(typeof(int));      // p is local_0, x is arg_0
    ilg.Emit(OpCodes.Ldc_I4_1);
    ilg.Emit(OpCodes.Stloc_0);          // p = 1;
    while (n > 0) {
      if (n % 2 == 0) { 
        ilg.Emit(OpCodes.Ldarg_0);      // x is arg_0
        ilg.Emit(OpCodes.Ldarg_0);
        ilg.Emit(OpCodes.Mul);
        ilg.Emit(OpCodes.Starg_S, 0);   // x = x * x
        n = n / 2; 
      } else { 
        ilg.Emit(OpCodes.Ldloc_0);
        ilg.Emit(OpCodes.Ldarg_0);
        ilg.Emit(OpCodes.Mul);
        ilg.Emit(OpCodes.Stloc_0);      // p = p * x; 
        n = n - 1; 
      }
    }
    ilg.Emit(OpCodes.Ldloc_0);
    ilg.Emit(OpCodes.Ret);              // return p;
  }

  public delegate int Int2Int(int x);
  public delegate int IntInt2Int(int x, int y);
}

public interface IIntInt2Int {
  int Power(int n, int x);
}

// Crude timing utility ----------------------------------------
   
public class Timer {
  private DateTime start;

  public Timer() {
    start = DateTime.Now;
  }

  public double Check() {
    TimeSpan dur = DateTime.Now - start;
    return dur.TotalSeconds;
  }
}
