// File RTCG5D.cs --- evaluation of polynomials, for CLR 2.0
// sestoft@dina.kvl.dk * 2002-10-02, 2005-04-23

// Mobile Pentium M 1.6GHz: The specialized code is faster only for
// polynomials of degree 20 with many zero coefficients.

using System;
using System.Reflection;
using System.Reflection.Emit;

class RTCG5D {
  public static void Main(String [] args) {
    int order = args.Length-1;
    double[] cs = new double[order];
    for (int i=0; i<order; i++)
      cs[i] = double.Parse(args[i]);
    int count = int.Parse(args[order]);

    // Build: public static double MyPolyEval(double x) { ... }
    DynamicMethod methodBuilder =
      new DynamicMethod("MyPolyEval", 
                        typeof(double),
                        new Type[] { typeof(double) },
                        typeof(String).Module);
    // Obtain an IL generator to build the method body
    ILGenerator ilg = methodBuilder.GetILGenerator();
    PolyGen(ilg, cs);

    D2D eval_cs = (D2D)methodBuilder.CreateDelegate(typeof(D2D));

    double eps = 2.0/count;
    double res;
    {
      res = 0.0;
      Timer t = new Timer();
      for (double x=0; x < 2; x += eps)
        res += eval_cs(x);
      Console.WriteLine("Delegate call to specialized generated method: " 
                        + t.Check() + " sec");
      Console.WriteLine(res);
    }
    res = 0.0;
    { 
      Timer t = new Timer();
      for (double x=0; x < 2; x += eps)
        res += Poly(cs, x);
      Console.WriteLine("Static call to general compiled method: " 
                        + t.Check() + " sec");
      Console.WriteLine(res);
    }
    {
      res = 0.0;
      Timer t = new Timer();
      for (double x=0; x < 2; x += eps)
        res += eval_cs(x);
      Console.WriteLine("Delegate call to specialized generated method: " 
                        + t.Check() + " sec");
      Console.WriteLine(res);
    }
    res = 0.0;
    { 
      Timer t = new Timer();
      for (double x=0; x < 2; x += eps)
        res += Poly(cs, x);
      Console.WriteLine("Static call to general compiled method: " 
                        + t.Check() + " sec");
      Console.WriteLine(res);
    }
  }

  // Evaluate polynomial cs[0] + cs[1] * x + cs[2] * x^2 + ...

  public static double Poly(double[] cs, double x) {
    double res = 0.0;
    for (int i=cs.Length-1; i>=0; i--) 
      res = res * x + cs[i];
    return res;
  }

  public static void PolyGen(ILGenerator ilg, double[] cs) { // x is arg_0
    ilg.Emit(OpCodes.Ldc_R8, 0.0);              // push res = 0.0 on stack
    for (int i=cs.Length-1; i>=0; i--) {
      ilg.Emit(OpCodes.Ldarg_0);                // load x
      ilg.Emit(OpCodes.Mul);                    // compute res * x
      if (cs[i] != 0.0) {
        ilg.Emit(OpCodes.Ldc_R8, cs[i]);        // load cs[i]
        ilg.Emit(OpCodes.Add);                  // compute x * res + cs[i]
      }
    }
    ilg.Emit(OpCodes.Ret);                      // return res;
  }

  public delegate double D2D(double x);
}

// Crude timing utility ----------------------------------------
   
public class Timer {
  private DateTime start;

  public Timer() {
    start = DateTime.Now;
  }

  public double Check() {
    TimeSpan dur = DateTime.Now - start;
    return dur.TotalSeconds;
  }
}
