// File RTCG7D.cs --- measure time to generate (much) code
// sestoft@dina.kvl.dk * 2002-09, 2005-10-05

using System;
using System.Reflection;
using System.Reflection.Emit;

class RTCG7 {
  public static void Main(String [] args) {
    if (args.Length != 2) {
      Console.WriteLine("Usage: RTCG7D <count> <calls>\n");
      return;
    }
    int count = int.Parse(args[0]);
    int calls = int.Parse(args[1]);
    // Build method: public static void MyMethod1(int x) { ... }
    DynamicMethod methodBuilder =
      new DynamicMethod("MyMethod", 
                        typeof(void),
                        new Type[] { typeof(int) },
                        typeof(String).Module);
    Timer t1 = new Timer();
    {
      // Obtain an IL generator to build the method body
      ILGenerator ilg = methodBuilder.GetILGenerator();
      ilg.Emit(OpCodes.Ldarg_0);
      ilg.Emit(OpCodes.Ldc_I4_1);
      ilg.Emit(OpCodes.Add);
      for (int i=count; i>0; i--) {
  	ilg.Emit(OpCodes.Ldarg_0);
  	ilg.Emit(OpCodes.Add);
      }
      ilg.Emit(OpCodes.Pop);  
      ilg.Emit(OpCodes.Ret);  
    }
    I2V mm = (I2V)methodBuilder.CreateDelegate(typeof(I2V));
    for (int i=calls; i>0; i--) {
      mm(count);
    }
    Console.WriteLine("Generation and " + calls + " calls: " 
		      + t1.Check() + " sec");
  }
}

public delegate void I2V(int x);

// Crude timing utility ----------------------------------------
   
public class Timer {
  private DateTime start;

  public Timer() {
    start = DateTime.Now;
  }

  public double Check() {
    TimeSpan dur = DateTime.Now - start;
    return dur.TotalSeconds;
  }
}
